/*
 * $Header: /src/NTP/REPOSITORY/v3/include/dcf77.h,v 3.7 1993/01/18 09:24:09 kardel Exp $
 *
 * $Id: dcf77.h,v 3.7 1993/01/18 09:24:09 kardel Exp $
 *
 * Copyright (c) 1989,1990,1991,1992,1993
 * Frank Kardel Friedrich-Alexander Universitaet Erlangen-Nuernberg
 *                                    
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Log: dcf77.h,v $
 * Revision 3.7  1993/01/18  09:24:09  kardel
 * updated copyright conditions in conjunction with
 * conditions set up in the COPYRIGHT file
 *
 * Revision 3.6  1993/01/17  13:42:44  kardel
 * 1993 initial update
 *
 */

#ifndef lint
  static char dcfhrcsid[]="$Id: dcf77.h,v 3.7 1993/01/18 09:24:09 kardel Exp $ FAU";
#endif

/*
 * we use the following datastructures in two modes
 * either in the NTP itself where we use NTP time stamps at some places
 * or in the kernel, where only struct timeval will be used.
 */
#ifdef KERNEL
#ifndef STREAM
#define STREAM
#endif
#endif
#ifdef STREAM
#define TIMESTAMP struct timeval
#define MALLOC(_X_) (char *)kmem_alloc(_X_)
#define FREE(_X_, _Y_) kmem_free((caddr_t)_X_, _Y_)
#else
#define TIMESTAMP l_fp
extern char *malloc();
#define MALLOC(_X_) malloc(_X_)
#define FREE(_X_, _Y_) free(_X_)
#endif

/*
 * state flags
 */
#define DCFB_ANNOUNCE           0x0001 /* switch time zone warning (DST switch) */
#define DCFB_POWERUP            0x0002 /* no synchronisation */
#define DCFB_NOSYNC             0x0004 /* timecode currently not confirmed */
#define DCFB_DST                0x0008 /* DST in effect */
#define DCFB_UTC		0x0010 /* UTC time */
#define DCFB_LEAP		0x0020 /* LEAP warning (1 hour prior to occurence) */
#define DCFB_ALTERNATE		0x0040 /* alternate antenna used */

#define DCFB_S_LEAP		0x0100 /* supports LEAP */
#define DCFB_S_LOCATION		0x0200 /* supports LOCATION information */
#define DCFB_S_PPS     		0x0400 /* supports PPS time stamping */

#define DCFB_TIMECODE		0x1000 /* valid time code sample */
#define DCFB_PPS		0x2000 /* valid PPS sample */

#define DCF_TCINFO		(DCFB_ANNOUNCE|DCFB_POWERUP|DCFB_NOSYNC|DCFB_DST|\
				 DCFB_UTC|DCFB_LEAP|DCFB_ALTERNATE|DCFB_S_LEAP|\
				 DCFB_S_LOCATION|DCFB_TIMECODE)

#define DCF_POWERUP(x)          ((x) & DCFB_POWERUP)
#define DCF_NOSYNC(x)           (((x) & (DCFB_POWERUP|DCFB_NOSYNC)) == DCFB_NOSYNC)
#define DCF_SYNC(x)             (((x) & (DCFB_POWERUP|DCFB_NOSYNC)) == 0)
#define DCF_ANNOUNCE(x)         ((x) & DCFB_ANNOUNCE)
#define DCF_DST(x)              ((x) & DCFB_DST)
#define DCF_UTC(x)		((x) & DCFB_UTC)
#define DCF_LEAP(x)		(DCF_SYNC(x) && ((x) & DCFB_LEAP))
#define DCF_ALTERNATE(x)	((x) & DCFB_ALTERNATE)
#define DCF_S_LEAP(x)		((x) & DCFB_S_LEAP)
#define DCF_S_LOCATION(x)	((x) & DCFB_S_LOCATION)
#define DCF_S_PPS(x)		((x) & DCFB_S_PPS)
#define DCF_TIMECODE(x)		((x) & DCFB_TIMECODE)
#define DCF_PPS(x)		((x) & DCFB_PPS)

/*
 * operation flags - some are also fudge flags
 */
#define DCF_STAT_FLAGS    0x03	/* interpreted by io module */
#define   DCF_STAT_FILTER 0x01	/* filter incoming data */
#define   DCF_STAT_AVG	  0x02	/* 1:median average / 0: median point */
#define DCF_FIXED_FMT     0x10  /* fixed format */
#define DCF_LEAP_DELETE   0x20	/* delete leap */

/*
 * standard time stamp structure
 */
struct dcftime
{
  unsigned long  dcf_status;	/* data status - CVT_OK, CVT_NONE, CVT_FAIL ... */
  struct timeval dcf_time;	/* DCF timestamp */
  TIMESTAMP	 dcf_stime;	/* telegram sample timestamp */
  struct timeval dcf_ptime;	/* PPS time stamp */
  long           dcf_usecerror;	/* sampled/filtered usec error */
  long           dcf_usecdisp;	/* sampled usecdispersion */
  unsigned long	 dcf_state;	/* current receiver state */
  unsigned short dcf_format;	/* format code */
};

typedef struct dcftime dcftime_t;

/*---------- STREAMS interface ----------*/

#ifdef STREAM
/*
 * ioctls
 */
#define DCFIOC_SETSTAT	(('D'<<8) + 'S')
#define DCFIOC_GETSTAT	(('D'<<8) + 'G')
#define DCFIOC_SETFMT   (('D'<<8) + 'f')
#define DCFIOC_GETFMT	(('D'<<8) + 'F')
#define DCFIOC_SETCS	(('D'<<8) + 'C')
#define DCFIOC_TIMECODE	(('D'<<8) + 'T')

#endif

/*------ IO handling flags (sorry) ------*/

#define DCF_IO_CSIZE	0x00000003
#define DCF_IO_CS5	0x00000000
#define DCF_IO_CS6	0x00000001
#define DCF_IO_CS7	0x00000002 
#define DCF_IO_CS8	0x00000003 

/*
 * sizes
 */
#define DCF_TCMAX	64

/*
 * ioctl structure
 */
union dcfctl 
{
  struct dcfstatus
    {
      unsigned long flags;	/* new/old flags */
    } dcfstatus;

  struct dcfgettc
    {
      unsigned long  dcf_state;	/* last state */
      unsigned long  dcf_badformat; /* number of bad packets since last query */
      unsigned short dcf_format;/* last decoded format */
      unsigned short dcf_count;	/* count of valid time code bytes */
      char           dcf_buffer[DCF_TCMAX+1]; /* timecode buffer */
    } dcfgettc;

  struct dcfformat
    {
      unsigned short dcf_format;/* number of examined format */
      unsigned short dcf_count;	/* count of valid string bytes */
      char           dcf_buffer[DCF_TCMAX+1]; /* format code string */
    } dcfformat;

  struct dcfsetcs
    {
      unsigned long  dcf_cs;	/* character size (needed for stripping) */
    } dcfsetcs;
};
  
typedef union dcfctl dcfctl_t;

/*------ for conversion routines --------*/

#define DCF_DELTA        26

struct dcf			/* dcf module local data */
{
  int            dcf_flags;	/* operation and current status flags */
  
  int		 dcf_ioflags;	   /* io handling flags (5-8 Bit control currently) */
  int		 dcf_syncflags;	   /* possible sync events (START/END/character) */
  /*
   * RS232 input parser information
   */
  char           dcf_startsym[32]; /* possible start packet values */
  char           dcf_endsym[32];   /* possible end packet values */
  char		 dcf_syncsym[32];  /* sync characters */

  /*
   * PPS 'input' buffer
   */
  struct timeval dcf_lastone;	/* time stamp of last PPS 1 transition */
  struct timeval dcf_lastzero;	/* time stamp of last PPS 0 transition */

  /*
   * time code input buffer (from RS232 or PPS)
   */
  int            dcf_index;	/* current buffer index */
  char          *dcf_data;      /* data buffer */
  unsigned short dcf_dsize;	/* size of data buffer */
  unsigned short dcf_lformat;	/* last format used */
  unsigned long  dcf_lstate;	/* last state code */
  char          *dcf_ldata;	/* last data buffer */
  unsigned short dcf_ldsize;	/* last data buffer length */
  unsigned long  dcf_badformat;	/* number of unparsable pakets */
  
  /*
   * time stamp filtering
   */
  long           dcf_delta[DCF_DELTA]; /* delta buffer */
  int            dcf_dindex;

  dcftime_t      dcf_dtime;	/* external data prototype */
};

typedef struct dcf dcf_t;

struct clocktime		/* clock time broken up from time code */
{
  long day;
  long month;
  long year;
  long hour;
  long minute;
  long second;
  long usecond;
  long utcoffset;	/* in minutes */
  long flags;		/* current clock status */
};

typedef struct clocktime clocktime_t;

/*
 * clock formats specify routines to be called to
 * convert the buffer into a struct clock.
 * functions are called
 *   fn(buffer, data, clock) -> CVT_NONE, CVT_FAIL, CVT_OK
 *
 * the private data pointer can be used to
 * distingush between different formats of a common
 * base type
 */
#define F_START		0x00000001 /* start packet delimiter */
#define F_END		0x00000002 /* end packet delimiter */
#define SYNC_START	0x00000004 /* packet start is sync event */
#define SYNC_END	0x00000008 /* packet end is sync event */
#define SYNC_CHAR	0x00000010 /* special character is sync event */
#define SYNC_ONE	0x00000020 /* PPS synchronize on 'ONE' transition */
#define SYNC_ZERO	0x00000040 /* PPS synchronize on 'ZERO' transition */
/*
 * parser related return/error codes
 */
#define CVT_MASK	0x0000000F /* conversion exit code */
#define   CVT_NONE	0x00000001 /* format not applicable */
#define   CVT_FAIL	0x00000002 /* conversion failed - error code returned */
#define   CVT_OK	0x00000004 /* conversion succeeded */
#define CVT_BADFMT	0x00000010 /* general format error - (unparsable) */
#define CVT_BADDATE     0x00000020 /* date field incorrect */
#define CVT_BADTIME	0x00000040 /* time field incorrect */

struct clockformat
{
  unsigned long (*convert)();	/* conversion routine */
  void          (*syncevt)();	/* routine for handling RS232 sync events (time stamps) */
  unsigned long (*syncpps)();	/* PPS input routine */
  void           *data;		/* local parameters */
  char           *name;		/* clock format name */
  unsigned short  length;	/* maximum length of data packet */
  unsigned long   flags;	/* valid start symbols etc. */
  char            startsym;	/* start symbol */
  char            endsym;	/* end symbol */
  char            syncsym;	/* sync symbol */
};

typedef struct clockformat clockformat_t;

/*
 * dcf77parse interface
 */
extern int  dcf_ioinit(/* dcf_t *dcfio */);
extern void dcf_ioend(/* dcf_t *dcfio */);
extern int  dcf_ioread(/* dcf_t *dcfio, char ch, TIMESTAMP *ctime, dcftime_t *dtime */);
extern int  dcf_iopps(/* dcf_t *dcfio, int status, struct timeval *ptime, dcftime_t *dtime */);
extern void dcf_iodone(/* dcf_t *dcfio */);

extern int  dcf_getstat(/* dcfctl_t *dct, dcf_t *dcf */);
extern int  dcf_setstat(/* dcfctl_t *dct, dcf_t *dcf */);
extern int  dcf_timecode(/* dcfctl_t *dct, dcf_t *dcf */);
extern int  dcf_getfmt(/* dcfctl_t *dct, dcf_t *dcf */);
extern int  dcf_setfmt(/* dcfctl_t *dct, dcf_t *dcf */);
extern int  dcf_setcs(/* dcfctl_t *dct, dcf_t *dcf */);

