#if defined(REFCLOCK) && defined(DCF77)
/*
 * $Header: /src/NTP/REPOSITORY/v3/lib/dcf77parse.c,v 3.13 1993/01/19 18:12:41 kardel Exp $
 *  
 * $Id: dcf77parse.c,v 3.13 1993/01/19 18:12:41 kardel Exp $
 *
 * Parser module for reference clock
 *
 * STREAM define switches between two personalities of the module
 * if STREAM is defined this module can be used with dcf77sync.c as
 * a STREAMS kernel module. In this case the time stamps will be
 * a struct timeval.
 * when STREAM is not defined NTP time stamps will be used.
 *
 * Copyright (c) 1992,1993
 * Frank Kardel Friedrich-Alexander Universitaet Erlangen-Nuernberg
 *                                    
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Log: dcf77parse.c,v $
 * Revision 3.13  1993/01/19  18:12:41  kardel
 * DCF77/PPS Release
 *
 */

#ifndef lint
static char rcsid[] = "$Id: dcf77parse.c,v 3.13 1993/01/19 18:12:41 kardel Exp $";
#endif

#include "sys/types.h"
#include "sys/time.h"
#include "sys/errno.h"

#ifdef STREAM
#include "sys/conf.h"
#include "sys/buf.h"
#include "sys/param.h"
#include "sys/sysmacros.h"
#include "sundev/mbvar.h"
#include "sun/autoconf.h"
#include "sys/stream.h"
#include "sys/stropts.h"
#else
#include "syslog.h"
#ifdef DEBUG
extern int debug;
#define DD_PARSE 5
#define dcfprintf(LEVEL, ARGS) if (debug > LEVEL) printf ARGS
#else
#define dcfprintf(LEVEL, ARGS)
#endif
#include "ntp_fp.h"
#include "ntp_unixtime.h"
#include "syslog.h"
#endif

#include "ntp_calendar.h"
#include "dcf77.h"
#ifdef STREAM
#include "sys/dcf77sync.h"
#endif

#define TIMES10(_X_)	(((_X_) << 3) + ((_X_) << 1))

static unsigned long timepacket();

static unsigned long cvt_simple();
static unsigned long cvt_dcf7000();

static void          syn_simple();
static unsigned long pps_simple();

/*
 * field location structure (Meinberg clocks/simple format)
 */
#define O_DAY	0
#define O_MONTH 1
#define O_YEAR	2
#define O_HOUR	3
#define O_MIN	4
#define	O_SEC	5
#define O_WDAY	6
#define O_FLAGS 7
#define O_ZONE  8
#define O_COUNT (O_ZONE+1)

#define MBG_EXTENDED	0x00000001

/*
 * see below for field offsets
 */
static struct format
{
  struct foff
    {
      char offset;		/* offset into buffer */
      char length;		/* length of field */
    } field_offsets[O_COUNT];
  char *fixed_string;		/* string with must be chars (blanks = wildcards) */
  unsigned long flags;
} formats[] =
{
  {
    {
      { 3, 2}, { 6, 2}, { 9, 2}, { 18, 2}, { 21, 2}, { 24, 2}, { 14, 1}, { 27, 4}, { 29, 1 }
      
    },
    "\2D:  .  .  ;T: ;U:  .  .  ;    \3",
    0
  },
  {				/* special extended FAU Erlangen extended format */
    {
      { 1, 2}, { 4, 2}, { 7, 2}, { 14, 2}, { 17, 2}, { 20, 2}, { 11, 1}, { 25, 4}, { 27, 1 }
    },
    "\2  .  .  ;  ;   :  :  ;        \3",
    MBG_EXTENDED
  },
  {				/* ELV DCF7000 */
    {
      { 6, 2}, { 3, 2}, { 0, 2}, { 12, 2}, { 15, 2}, { 18, 2}, { 9, 2}, { 21, 2}, { 0, 0 }
    },
    "  -  -  -  -  -  -  -  \r",
    0
  }
};    

/*
 * format definitions
 */
static clockformat_t clockformats[] =
{
  {
    cvt_simple,			/* Meinberg conversion */
    syn_simple,			/* easy time stamps for RS232 (fallback) */
    pps_simple,			/* easy PPS monitoring */
    (void *)&formats[0],	/* conversion configuration */
    "Meinberg Standard",	/* Meinberg simple format - beware */
    32,				/* string buffer */
    F_START|F_END|SYNC_START|SYNC_ONE, /* paket START/END delimiter, START synchronisation, PPS ONE sampling */
    '\2',
    '\3',
    '\0'
  },
  {
    cvt_simple,			/* Meinberg conversion */
    syn_simple,			/* easy time stamps for RS232 (fallback) */
    pps_simple,			/* easy PPS monitoring */
    (void *)&formats[1],	/* conversion configuration */
    "Meinberg Extended",	/* Meinberg enhanced format */
    32,				/* string buffer */
    F_START|F_END|SYNC_START|SYNC_ONE,	/* paket START/END delimiter, START synchronisation, PPS ONE sampling  */
    '\2',
    '\3',
    '\0'
  },
  {
    cvt_dcf7000,		/* ELV DCF77 conversion */
    syn_simple,			/* easy time stamps */
    (unsigned long (*)())0,	/* no direct PPS monitoring */
    (void *)&formats[2],	/* conversion configuration */
    "ELV DCF7000",		/* ELV clock */
    24,				/* string buffer */
    F_END|SYNC_END,		/* END packet delimiter / synchronisation */
    '\0',
    '\r',
    '\0'
  }
};

static nformats = sizeof(clockformats) / sizeof(clockformats[0]);

#ifdef STREAM
/*
 * strings support usually not in kernel - duplicated, but what the heck
 */
static int strlen(s)
  register char *s;
{
  register int c;

  c = 0;
  if (s)
    {
      while (*s++)
	{
	  c++;
	}
    }
  return c;
}

static int strcmp(s, t)
  register char *s;
  register char *t;
{
  register int c = 0;

  if (!s || !t || (s == t))
    {
      return 0;
    }

  while (!(c = *s++ - *t++) && *s && *t)
    /* empty loop */;
  
  return c;
}
#endif

/*ARGSUSED*/
int dcf_ioinit(dcfio)
  register dcf_t *dcfio;
{
  register int i;
  register int f = 0;
  
  dcfprintf(DD_PARSE, ("dcf_iostart\n"));
  /*
   * gather bitmaps of possible start and end values
   */
  for (i=0; i < nformats; i++)
    {
      if (clockformats[i].flags & F_START)
	{
	  if (dcfio->dcf_endsym[clockformats[i].startsym / 8] & (1 << (clockformats[i].startsym % 8)))
	    {
	      dcfprintf(DD_PARSE, ("init failed: START symbol collides with END symbol (format %d)\n", i));
	      return 0;
	    }
	  else
	    {
	      dcfio->dcf_startsym[clockformats[i].startsym / 8] |= (1 << (clockformats[i].startsym % 8));
	      f = 1;
	    }
	}

      if (clockformats[i].flags & F_END)
	{
	  if (dcfio->dcf_startsym[clockformats[i].endsym / 8] & (1 << (clockformats[i].endsym % 8)))
	    {
	      dcfprintf(DD_PARSE, ("init failed: END symbol collides with START symbol (format %d)\n", i));
	      return 0;
	    }
	  else
	    {
	      dcfio->dcf_endsym[clockformats[i].endsym / 8] |= (1 << (clockformats[i].endsym % 8));
	      f = 1;
	    }
	}
      if (clockformats[i].flags & SYNC_CHAR)
	{
	  dcfio->dcf_syncsym[clockformats[i].syncsym / 8] |= (1 << (clockformats[i].syncsym % 8));
	}
      dcfio->dcf_syncflags |= clockformats[i].flags & (SYNC_START|SYNC_END|SYNC_CHAR|SYNC_ONE|SYNC_ZERO);
      if (dcfio->dcf_dsize < clockformats[i].length)
	dcfio->dcf_dsize = clockformats[i].length;
    }
  
  if (!f)
    {
      /*
       * need at least one start or end symbol
       */
      dcfprintf(DD_PARSE, ("init failed: neither START nor END symbol defined\n"));
      return 0;
    }

  dcfio->dcf_data = MALLOC(dcfio->dcf_dsize * 2 + 2);
  if (!dcfio->dcf_data)
    {
      dcfprintf(DD_PARSE, ("init failed: malloc for data area failed\n"));
      return 0;
    }

  /*
   * leave room for '\0'
   */
  dcfio->dcf_ldata = dcfio->dcf_data + dcfio->dcf_dsize + 1;
  dcfio->dcf_lformat = 0;
  dcfio->dcf_badformat = 0;
  dcfio->dcf_ioflags   = DCF_IO_CS7;	/* usual unix default */
  return 1;
}

/*ARGSUSED*/
void dcf_ioend(dcfio)
  register dcf_t *dcfio;
{
  dcfprintf(DD_PARSE, ("dcf_ioend\n"));
  if (dcfio->dcf_data)
    FREE(dcfio->dcf_data, dcfio->dcf_dsize * 2 + 2);
}

/*ARGSUSED*/
int dcf_ioread(dcfio, ch, ctime)
  register dcf_t *dcfio;
  register char ch;
  register TIMESTAMP *ctime;
{
  register unsigned updated = CVT_NONE;
  
  dcfprintf(DD_PARSE, ("dcf_ioread(0x%x, char=0x%x, ..., ...)\n", dcfio, ch & 0xFF));

  if ((dcfio->dcf_flags & DCF_FIXED_FMT)
      && !clockformats[dcfio->dcf_lformat].convert)
    {
      dcfprintf(DD_PARSE, ("dcf_ioread: input dropped.\n"));
      return CVT_NONE;
    }
		  
  /*
   * within STREAMS CSx (x < 8) chars still have the upper bits set
   * so we normalize the characters by masking unecessary bits off.
   */
  switch (dcfio->dcf_ioflags & DCF_IO_CSIZE)
    {
    case DCF_IO_CS5:
      ch &= 0x1F;
      break;

    case DCF_IO_CS6:
      ch &= 0x3F;
      break;

    case DCF_IO_CS7:
      ch &= 0x7F;
      break;
      
    case DCF_IO_CS8:
      break;
    }

  if ((dcfio->dcf_syncflags & SYNC_CHAR) &&
      (dcfio->dcf_syncsym[ch / 8] & (1 << (ch % 8))))
    {
      register int i;
      /*
       * got a sync event - call sync routine
       */

      for (i = 0; i < nformats; i++)
	{
	  if ((clockformats[i].flags & SYNC_CHAR) &&
	      (clockformats[i].syncsym == ch))
	    {
	      dcfprintf(DD_PARSE, ("dcf_ioread: SYNC_CHAR event\n"));
	      if (clockformats[i].syncevt)
		clockformats[i].syncevt(dcfio, ctime, clockformats[i].data, SYNC_CHAR);
	      else
		{
#ifdef STREAM
		  printf("dcf77: dcf_ioread: WARNING: SYNC_CHAR event attempt for format %d (\"%s\") event character 0x%x\n", i, clockformats[i].name, ch);
#else
		  syslog(LOG_WARNING, "dcf77: dcf_ioread: WARNING: SYNC_CHAR event attempt for format %d (\"%s\") event character 0x%x\n", i, clockformats[i].name, ch);
#endif
		}
	    }
	}
    }
		
  if (dcfio->dcf_startsym[ch / 8] & (1 << (ch % 8))) 
    {
      register int i;
      /*
       * packet start - re-fill buffer
       */
      if (dcfio->dcf_index)
	{
	  /*
	   * filled buffer - thus not end character found
	   * do processing now
	   */
	  dcfio->dcf_data[dcfio->dcf_index] = '\0';
	  bcopy(dcfio->dcf_data, dcfio->dcf_ldata, dcfio->dcf_index+1);
	  dcfio->dcf_ldsize = dcfio->dcf_index+1;
			
	  updated = timepacket(dcfio);
	}

      /*
       * could be a sync event - call sync routine if needed
       */
      if (dcfio->dcf_syncflags & SYNC_START)
	for (i = 0; i < nformats; i++)
	  {
	    if ((clockformats[i].flags & SYNC_START) && (clockformats[i].startsym == ch))
	      {
		dcfprintf(DD_PARSE, ("dcf_ioread: SYNC_START event\n"));
		if (clockformats[i].syncevt)
		  clockformats[i].syncevt(dcfio, ctime, clockformats[i].data, SYNC_START);
		else
		  {
#ifdef STREAM
		    printf("dcf77: dcf_ioread: WARNING: SYNC_START event attempt for format %d (\"%s\") event character 0x%x\n", i, clockformats[i].name, ch);
#else
		    syslog(LOG_WARNING, "dcf77: dcf_ioread: WARNING: SYNC_START event attempt for format %d (\"%s\") event character 0x%x\n", i, clockformats[i].name, ch);
#endif
		  }
	      }
	  }
      dcfio->dcf_index = 1;
      dcfio->dcf_data[0] = ch;
    }
  else
    {
      register int i;

      if (dcfio->dcf_index < dcfio->dcf_dsize)
	{
	  /*
	   * collect into buffer
	   */
	  dcfio->dcf_data[dcfio->dcf_index++] = ch;
	}
      if ((dcfio->dcf_endsym[ch / 8] & (1 << (ch % 8))) ||
	  (dcfio->dcf_index >= dcfio->dcf_dsize))
	{
	  /*
	   * packet end - process buffer
	   */
	  if (dcfio->dcf_syncflags & SYNC_END)
	    for (i = 0; i < nformats; i++)
	      {
		if ((clockformats[i].flags & SYNC_END) && (clockformats[i].endsym == ch))
		  {
		    dcfprintf(DD_PARSE, ("dcf_ioread: SYNC_END event\n"));
		    if (clockformats[i].syncevt)
		      clockformats[i].syncevt(dcfio, ctime, clockformats[i].data, SYNC_END);
		    else
		      {
#ifdef STREAM
			printf("dcf77: dcf_ioread: WARNING: SYNC_END event attempt for format %d (\"%s\") event character 0x%x\n", i, clockformats[i].name, ch);
#else
			syslog(LOG_WARNING, "dcf77: dcf_ioread: WARNING: SYNC_END event attempt for format %d (\"%s\") event character 0x%x\n", i, clockformats[i].name, ch);
#endif
		      }
		  }
	      }
	  dcfio->dcf_data[dcfio->dcf_index] = '\0';
	  bcopy(dcfio->dcf_data, dcfio->dcf_ldata, dcfio->dcf_index+1);
	  dcfio->dcf_ldsize = dcfio->dcf_index+1;
		    
	  updated = timepacket(dcfio);
	  dcfio->dcf_index = 0;
	}
    }
#ifdef DEBUG
  if ((updated & CVT_MASK) != CVT_NONE)
    dcfprintf(DD_PARSE, ("dcf_ioread: time sample accumulated (status=0x%x)\n", updated));
#endif

  dcfio->dcf_dtime.dcf_status = updated;

  return (updated & CVT_MASK) != CVT_NONE;
}

/*
 * dcf_iopps
 *
 * take status line indication and derive synchronisation information
 * from it.
 * It can also be used to decode a serial serial data format (such as the
 * ONE, ZERO, MINUTE sync data stream from DCF77)
 */
/*ARGSUSED*/
int dcf_iopps(dcfio, status, ptime)
  register dcf_t *dcfio;
  register int status;
  register struct timeval *ptime;
{
  register unsigned updated = CVT_NONE;

  /*
   * PPS pulse information will only be delivered to ONE clock format
   * this is either the last successful conversion module with a ppssync
   * routine, or a fixed format with a ppssync routine
   */
  dcfprintf(DD_PARSE, ("dcf_iopps: STATUS %s\n", (status == SYNC_ONE) ? "ONE" : "ZERO"));

  if (((dcfio->dcf_flags & DCF_FIXED_FMT) ||
       ((dcfio->dcf_dtime.dcf_status & CVT_MASK) == CVT_OK)) &&
      clockformats[dcfio->dcf_lformat].syncpps &&
      (status & clockformats[dcfio->dcf_lformat].flags))
    {
      updated = clockformats[dcfio->dcf_lformat].syncpps(dcfio, status == SYNC_ONE, ptime);
      dcfprintf(DD_PARSE, ("dcf_iopps: updated = 0x%x\n", updated));
    }
  else
    {
      dcfprintf(DD_PARSE, ("dcf_iopps: STATUS dropped\n"));
    }

  return (updated & CVT_MASK) != CVT_NONE;
}

/*
 * dcf_iodone
 *
 * clean up internal status for new round
 */
/*ARGSUSED*/
void dcf_iodone(dcfio)
  register dcf_t *dcfio;
{
  /*
   * we need to clean up certain flags for the next round
   */
  dcfio->dcf_dtime.dcf_state = 0; /* no problems with ISRs */
}

/*---------- conversion implementation --------------------*/

/*
 * convert a struct clock to UTC since Jan, 1st 1970 0:00 (the UNIX EPOCH)
 */
#define dysize(x)	((x) % 4 ? 365 : ((x % 400) ? 365 :366))

static time_t dcf_to_unixtime(clock, cvtrtc)
  register clocktime_t   *clock;
  register unsigned long *cvtrtc;
{
#define SETRTC(_X_)	{ if (cvtrtc) *cvtrtc = (_X_); }
  static int days_of_month[] = 
    {
      0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
    };
  register int i;
  time_t t;
  
  if (clock->year < 100)
    clock->year += 1900;

  if (clock->year < 1970)
    clock->year += 100;		/* XXX this will do it till <2070 */

  if (clock->year < 0)
    {
      SETRTC(CVT_FAIL|CVT_BADDATE);
      return -1;
    }
  
  /*
   * sorry, slow section here - but it's not time critical anyway
   */
  t =  (clock->year - 1970) * 365;
  t += (clock->year >> 2) - (1970 >> 2);
  t -= clock->year / 400 - 1970 / 400;

  				/* month */
  if (clock->month <= 0 || clock->month > 12)
    {
      SETRTC(CVT_FAIL|CVT_BADDATE);
      return -1;		/* bad month */
    }
				/* adjust leap year */
  if (clock->month >= 3 && dysize(clock->year) == 366)
      t++;

  for (i = 1; i < clock->month; i++)
    {
      t += days_of_month[i];
    }
				/* day */
  if (clock->day < 1 || ((clock->month == 2 && dysize(clock->year) == 366) ?
		  clock->day > 29 : clock->day > days_of_month[clock->month]))
    {
      SETRTC(CVT_FAIL|CVT_BADDATE);
      return -1;		/* bad day */
    }

  t += clock->day - 1;
				/* hour */
  if (clock->hour < 0 || clock->hour >= 24)
    {
      SETRTC(CVT_FAIL|CVT_BADTIME);
      return -1;		/* bad hour */
    }

  t = TIMES24(t) + clock->hour;

  				/* min */
  if (clock->minute < 0 || clock->minute > 59)
    {
      SETRTC(CVT_FAIL|CVT_BADTIME);
      return -1;		/* bad min */
    }

  t = TIMES60(t) + clock->minute;
				/* sec */
  
  t += clock->utcoffset;	/* warp to UTC */

  if (clock->second < 0 || clock->second > 60)	/* allow for LEAPs */
    {
      SETRTC(CVT_FAIL|CVT_BADTIME);
      return -1;		/* bad sec */
    }

  t  = TIMES60(t) + clock->second;
				/* done */
  return t;
}

/*--------------- format conversion -----------------------------------*/

static int stoi(s, zp, cnt)
  char *s;
  long *zp;
  int cnt;
{
  char *b = s;
  int f,z,v;
  char c;

  f=z=v=0;

  while(*s == ' ')
    s++;
  
  if (*s == '-')
    {
      s++;
      v = 1;
    }
  else
    if (*s == '+')
      s++;
  
  for(;;)
    {
      c = *s++;
      if (c == '\0' || c < '0' || c > '9' || (cnt && ((s-b) > cnt)))
	{
	  if (f == 0)
	    {
	      return(-1);
	    }
	  if (v)
	    z = -z;
	  *zp = z;
	  return(0);
	}
      z = (z << 3) + (z << 1) + ( c - '0' );
      f=1;
    }
}


static strok(s, m)
  char *s;
  char *m;
{
  if (!s || !m)
    return 0;

  while(*s && *m)
    {
      if ((*m == ' ') ? 1 : (*s == *m))
	{
	  s++;
	  m++;
	}
      else
	{
	  return 0;
	}
    }
  return !*m;
}

/*
 * The Meinberg receiver every second sends a datagram of the following form
 * (Standard Format)
 * 
 *     <STX>D:<dd>.<mm>.<yy>;T:<w>;U:<hh>:<mm>:<ss>;<S><F><D><A><ETX>
 * pos:  0  00 00 0 00 0 11 111 1 111 12 2 22 2 22 2 2  2  3  3   3
 *       1  23 45 6 78 9 01 234 5 678 90 1 23 4 56 7 8  9  0  1   2
 * <STX>           = '\002' ASCII start of text
 * <ETX>           = '\003' ASCII end of text
 * <dd>,<mm>,<yy>  = day, month, year(2 digits!!)
 * <w>             = day of week (sunday= 0)
 * <hh>,<mm>,<ss>  = hour, minute, second
 * <S>             = '#' if never synced since powerup else ' ' for DCF U/A 31
 *                   '#' if not PZF sychronisation available else ' ' for PZF 535
 * <F>             = '*' if time comes from internal quartz else ' '
 * <D>             = 'S' if daylight saving time is active else ' '
 * <A>             = '!' during the hour preceeding an daylight saving time
 *                       start/end change
 *
 * For the university of Erlangen a special format was implemented to support
 * LEAP announcement and anouncement of alternate antenna.
 *
 * Version for UNI-ERLANGEN Software is: PZFUERL V4.6 (Meinberg)
 *
 * The use of this software release (or higher) is *ABSOLUTELY*
 * recommended due to the LEAP second support and UTC indication.
 * The standard timecode does not indicate when the timecode is in
 * UTC (by front panel configuration) thus we hav no chance to find
 * the correct utc offset. For the standard format do not ever use
 * UTC display as this is no detectable in the time code !!!
 *
 *     <STX><dd>.<mm>.<yy>; <w>; <hh>:<mm>:<ss>; <U><S><F><D><A><L><R><ETX>
 * pos:  0   00 0 00 0 00 11 1 11 11 1 11 2 22 22 2  2  2  2  2  3  3   3
 *       1   23 4 56 7 89 01 2 34 56 7 89 0 12 34 5  6  7  8  9  0  1   2
 * <STX>           = '\002' ASCII start of text
 * <ETX>           = '\003' ASCII end of text
 * <dd>,<mm>,<yy>  = day, month, year(2 digits!!)
 * <w>             = day of week (sunday= 0)
 * <hh>,<mm>,<ss>  = hour, minute, second
 * <S>             = '#' if never synced since powerup else ' ' for DCF U/A 31
 *                   '#' if not PZF sychronisation available else ' ' for PZF 535
 * <U>             = 'U' UTC time display
 * <F>             = '*' if time comes from internal quartz else ' '
 * <D>             = 'S' if daylight saving time is active else ' '
 * <A>             = '!' during the hour preceeding an daylight saving time
 *                       start/end change
 * <L>             = 'A' LEAP second announcement
 * <R>             = 'R' alternate antenna
 */

/*
 * cvt_simple
 *
 * convert simple type format
 */
static unsigned long cvt_simple(buffer, format, clock)
  register char          *buffer;
  register struct format *format;
  register clocktime_t   *clock;
{
  if (!strok(buffer, format->fixed_string))
    {
      return CVT_NONE;
    }
  else
    {
      if (stoi(&buffer[format->field_offsets[O_DAY].offset], &clock->day,
	       format->field_offsets[O_DAY].length) ||
	  stoi(&buffer[format->field_offsets[O_MONTH].offset], &clock->month,
	       format->field_offsets[O_MONTH].length) ||
	  stoi(&buffer[format->field_offsets[O_YEAR].offset], &clock->year,
	       format->field_offsets[O_YEAR].length) ||
	  stoi(&buffer[format->field_offsets[O_HOUR].offset], &clock->hour,
	       format->field_offsets[O_HOUR].length) ||
	  stoi(&buffer[format->field_offsets[O_MIN].offset], &clock->minute,
	       format->field_offsets[O_MIN].length) ||
	  stoi(&buffer[format->field_offsets[O_SEC].offset], &clock->second,
	       format->field_offsets[O_SEC].length))
	{
	  return CVT_FAIL|CVT_BADFMT;
	}
      else
	{
	  char *f = &buffer[format->field_offsets[O_FLAGS].offset];
	  
	  clock->flags = 0;
	  clock->usecond = 0;

	  /*
	   * in the extended timecode format we have also the
	   * indication that the timecode is in UTC
	   * for compatibilty reasons we start at the USUAL
	   * offset (POWERUP flag) and know that the UTC indication
	   * is the character before the powerup flag
	   */
	  if ((format->flags & MBG_EXTENDED) && (f[-1] == 'U'))
	    {
	      /*
	       * timecode is in UTC
	       */
	      clock->utcoffset = 0; /* UTC */
	      clock->flags    |= DCFB_UTC;
	    }
	  else
	    {
	      /*
	       * only calculate UTC offset if MET/MED is in time code
	       * or we have the old time code format, where we do not
	       * know whether it is UTC time or MET/MED
	       * pray that nobody switches to UTC in the standard time code
	       * ROMS !!!!
	       */
	      switch (buffer[format->field_offsets[O_ZONE].offset])
		{
		case ' ':
		  clock->utcoffset = -60; /* MET */
		  break;
		  
		case 'S':
		  clock->utcoffset = -120; /* MED */
		  break;
		  
		default:
		  return CVT_FAIL|CVT_BADFMT;
		}
	    }
	  
	  /*
	   * gather status flags
	   */
	  if (buffer[format->field_offsets[O_ZONE].offset] == 'S')
	    clock->flags    |= DCFB_DST;
	  
	  if (f[0] == '#')
	    clock->flags |= DCFB_POWERUP;

	  if (f[1] == '*')
	    clock->flags |= DCFB_NOSYNC;

	  if (f[3] == '!')
	    clock->flags |= DCFB_ANNOUNCE;
	  
	  if (format->flags & MBG_EXTENDED)
	    {
	      clock->flags |= DCFB_S_LEAP;
	      clock->flags |= DCFB_S_LOCATION;
	      
	      if (f[4] == 'A')
		clock->flags |= DCFB_LEAP;

	      if (f[5] == 'R')
		clock->flags |= DCFB_ALTERNATE;
	    }
	  return CVT_OK;
	}
    }
}

/*
 * cvt_dcf7000
 *
 * convert dcf7000 type format
 */
static unsigned long cvt_dcf7000(buffer, format, clock)
  register char          *buffer;
  register struct format *format;
  register clocktime_t   *clock;
{
  if (!strok(buffer, format->fixed_string))
    {
      return CVT_NONE;
    }
  else
    {
      if (stoi(&buffer[format->field_offsets[O_DAY].offset], &clock->day,
	       format->field_offsets[O_DAY].length) ||
	  stoi(&buffer[format->field_offsets[O_MONTH].offset], &clock->month,
	       format->field_offsets[O_MONTH].length) ||
	  stoi(&buffer[format->field_offsets[O_YEAR].offset], &clock->year,
	       format->field_offsets[O_YEAR].length) ||
	  stoi(&buffer[format->field_offsets[O_HOUR].offset], &clock->hour,
	       format->field_offsets[O_HOUR].length) ||
	  stoi(&buffer[format->field_offsets[O_MIN].offset], &clock->minute,
	       format->field_offsets[O_MIN].length) ||
	  stoi(&buffer[format->field_offsets[O_SEC].offset], &clock->second,
	       format->field_offsets[O_SEC].length))
	{
	  return CVT_FAIL|CVT_BADFMT;
	}
      else
	{
	  char *f = &buffer[format->field_offsets[O_FLAGS].offset];
	  long flags;
	  
	  clock->flags = 0;
	  clock->usecond = 0;

	  if (stoi(f, &flags, format->field_offsets[O_FLAGS].length))
	    {
	      return CVT_FAIL|CVT_BADFMT;
	    }
	  else
	    {
	      if (flags & 0x1)
		clock->utcoffset = -120;
	      else
		clock->utcoffset = -60;

	      if (flags & 0x2)
		clock->flags |= DCFB_ANNOUNCE;

	      if (flags & 0x4)
		clock->flags |= DCFB_NOSYNC;
	    }
	  return CVT_OK;
	}
    }
}

/*
 * syn_simple
 *
 * handle a sync time stamp
 */
/*ARGSUSED*/
static void syn_simple(dcfio, ts, format, why)
  register dcf_t *dcfio;
  register TIMESTAMP *ts;
  register struct format *format;
  register unsigned long why;
{
  dcfio->dcf_dtime.dcf_stime = *ts;
}

/*
 * pps_simple
 *
 * handle a pps time stamp
 */
/*ARGSUSED*/
static unsigned long pps_simple(dcfio, status, ptime)
  register dcf_t *dcfio;
  register int status;
  register struct timeval *ptime;
{
  dcfio->dcf_dtime.dcf_ptime  = *ptime;
  dcfio->dcf_dtime.dcf_state |= DCFB_PPS|DCFB_S_PPS;
  return CVT_NONE;
}

/*
 * timepacket
 *
 * process a data packet
 */
static unsigned long timepacket(dcfio)
  register dcf_t *dcfio;
{
  register int s, k;
  register unsigned short format;
  register long mean;
  register time_t t;
  register long usecoff;
  register unsigned long cvtsum = 0;/* accumulated CVT_FAIL errors */
  unsigned long cvtrtc;		/* current conversion result */
  clocktime_t clock;
  long delta[DCF_DELTA];
  
  format = dcfio->dcf_lformat;

  k = 0;

  if (dcfio->dcf_flags & DCF_FIXED_FMT)
    {
      switch ((cvtrtc = clockformats[format].convert ? clockformats[format].convert(dcfio->dcf_data, clockformats[format].data, &clock) : CVT_NONE) & CVT_MASK)
	{
	case CVT_FAIL:
	  dcfio->dcf_badformat++;
	  cvtsum = cvtrtc & ~CVT_MASK;
	      
	  /*
	   * may be too often ... but is nice to know when it happens
	   */
#ifdef STREAM
	  printf("dcf77: \"%s\" failed to convert\n", clockformats[format].name);
#else
	  syslog(LOG_WARNING, "dcf77: \"%s\" failed to convert\n", clockformats[format].name);
#endif
	  break;
	  
	case CVT_NONE:
	  /*
	   * too bad - pretend bad format
	   */
	  dcfio->dcf_badformat++;
	  cvtsum = CVT_BADFMT;

	  break;
	  
	case CVT_OK:
	  k = 1;
	  break;

	default:
	  /* shouldn't happen */
#ifdef STREAM
	  printf("dcf77: INTERNAL error: bad return code of convert routine \"%s\"\n", clockformats[format].name);
#else
	  syslog(LOG_WARNING, "dcf77: INTERNAL error: bad return code of convert routine \"%s\"\n", clockformats[format].name);
#endif	  
	  return CVT_FAIL;
	}
    }
  else
    {
      /*
       * find correct conversion routine
       * and convert time packet
       * RR search starting at last successful conversion routine
       */
  
      if (nformats)		/* very careful ... */
	{
	  do
	    {
	      switch ((cvtrtc = clockformats[format].convert ? clockformats[format].convert(dcfio->dcf_data, clockformats[format].data, &clock) : CVT_NONE) & CVT_MASK)
		{
		case CVT_FAIL:
		  dcfio->dcf_badformat++;
		  cvtsum |= cvtrtc & ~CVT_MASK;
	      
		  /*
		   * may be too often ... but is nice to know when it happens
		   */
#ifdef STREAM
		  printf("dcf77: \"%s\" failed to convert\n", clockformats[format].name);
#else		  
		  syslog(LOG_WARNING, "dcf77: \"%s\" failed to convert\n", clockformats[format].name);
#endif
		  /*FALLTHROUGH*/
		case CVT_NONE:
		  format++;
		  break;
	  
		case CVT_OK:
		  k = 1;
		  break;

		default:
		  /* shouldn't happen */
#ifdef STREAM
		  printf("dcf77: INTERNAL error: bad return code of convert routine \"%s\"\n", clockformats[format].name);
#else
		  syslog(LOG_WARNING, "dcf77: INTERNAL error: bad return code of convert routine \"%s\"\n", clockformats[format].name);
#endif
		  return CVT_FAIL;
		}
	      if (format >= nformats)
		format = 0;
	    }
	  while (!k && (format != dcfio->dcf_lformat));
	}
  
    }

  if (!k)
    {
#ifdef STREAM
      printf("dcf77: time format \"%s\" not convertable\n", dcfio->dcf_data);
#else
      syslog(LOG_WARNING, "dcf77: time format \"%s\" not convertable\n", dcfio->dcf_data);
#endif
      return CVT_FAIL|cvtsum;
    }
  
  
  if ((t = dcf_to_unixtime(&clock, &cvtrtc)) == -1)
    {
#ifdef STREAM
      printf("dcf77: bad time format \"%s\"\n", dcfio->dcf_data);
#else
      syslog(LOG_WARNING,"dcf77: bad time format \"%s\"\n", dcfio->dcf_data);
#endif
      return CVT_FAIL|cvtrtc;
    }
  
  dcfio->dcf_lformat = format;
  
  {
#ifdef STREAM
    usecoff = (t - dcfio->dcf_dtime.dcf_stime.tv_sec) * 1000000
      - dcfio->dcf_dtime.dcf_stime.tv_usec + clock.usecond;
#else
    extern long tstouslo[];
    extern long tstousmid[];
    extern long tstoushi[];

    TSFTOTVU(usecoff, dcfio->dcf_stime.l_uf);
    usecoff  = -usecoff;
    usecoff += (t - (dcfio->dcf_stime.l_ui - JAN_1970)) * 1000000
              + clock->usecond;
#endif
  }
  
  /*
   * filtering (median) if requested
   */
  if (dcfio->dcf_flags & DCF_STAT_FILTER)
    {
      register int n, i;
      
      dcfio->dcf_delta[dcfio->dcf_dindex] = usecoff;

      dcfio->dcf_dindex = (dcfio->dcf_dindex + 1) % DCF_DELTA;

      /*
       * sort always - thus every sample gets its data
       */
      bcopy((caddr_t)dcfio->dcf_delta, (caddr_t)delta, sizeof(delta));
  
      for (s = 0; s < DCF_DELTA; s++)
	for (k = s+1; k < DCF_DELTA; k++)
	  {			/* Yes - it's slow sort */
	    if (delta[s] > delta[k]) 
	      {
		register long tmp;
		
		tmp      = delta[k];
		delta[k] = delta[s];
		delta[s] = tmp;
	      }
	  }

      i = 0;
      n = DCF_DELTA;

      /*
       * you know this median loop if you have read the other code
       */
      while ((n - i) > 8)
	{
	  register long top = delta[n-1];
	  register long mid = delta[(n+i)>>1];
	  register long low = delta[i];
	  
	  if ((top - mid) > (mid - low))
	    {
	      /*
	       * cut off high end
	       */
	      n--;
	    }
	  else
	    {
	      /*
	       * cut off low end
	       */
	      i++;
	    }
	}
      
      dcfio->dcf_dtime.dcf_usecdisp  = delta[n-1] - delta[i];

      if (dcfio->dcf_flags & DCF_STAT_AVG)
	{
	  /*
	   * take the average of the median samples as this clock
	   * is a little bumpy
	   */
	  mean = 0;

	  while (i < n)
	    {
	      mean += delta[i++];
	    }

	  mean >>= 3;
	}
      else
	{
	  mean = delta[(n+i)>>1];
	}
      
      dcfio->dcf_dtime.dcf_usecerror = mean;
    }
  else
    {
      dcfio->dcf_dtime.dcf_usecerror = usecoff;
      dcfio->dcf_dtime.dcf_usecdisp  = 0;
    }
  

  dcfprintf(DD_PARSE,("dcf77:timepacket: T=%x+%d usec, useccoff=%d, usecerror=%d, usecdisp=%d\n",
		      t, clock.usecond, usecoff, dcfio->dcf_dtime.dcf_usecerror, dcfio->dcf_dtime.dcf_usecdisp));
  
  /*
   * time stamp
   */
  dcfio->dcf_dtime.dcf_time.tv_sec  = t;
  dcfio->dcf_dtime.dcf_time.tv_usec = clock.usecond;
  dcfio->dcf_dtime.dcf_format       = format;

#ifdef STREAM
  {
    int s = splhigh();
#endif
  
    dcfio->dcf_lstate          = dcfio->dcf_dtime.dcf_state |
                                  clock.flags | DCFB_TIMECODE;
    
    dcfio->dcf_dtime.dcf_state = dcfio->dcf_lstate;

#ifdef STREAM
    (void)splx(s);
  }
#endif
  
  return CVT_OK;		/* everything fine and dandy... */
}

/*
 * control operations
 */
/*ARGSUSED*/
int dcf_getstat(dct, dcf)
  dcfctl_t *dct;
  dcf_t    *dcf;
{
  dct->dcfstatus.flags = dcf->dcf_flags & DCF_STAT_FLAGS;
  return 1;
}

		  
/*ARGSUSED*/
int dcf_setstat(dct, dcf)
  dcfctl_t *dct;
  dcf_t    *dcf;
{
  dcf->dcf_flags = (dcf->dcf_flags & ~DCF_STAT_FLAGS) | dct->dcfstatus.flags;
  return 1;
}

		  
/*ARGSUSED*/
int dcf_timecode(dct, dcf)
  dcfctl_t *dct;
  dcf_t    *dcf;
{
  dct->dcfgettc.dcf_state  = dcf->dcf_lstate;
  dct->dcfgettc.dcf_format = dcf->dcf_lformat;
  /*
   * move out current bad packet count
   * user program is expected to sum these up
   * this is not a problem, as "dcf" module are
   * exclusive open only
   */
  dct->dcfgettc.dcf_badformat = dcf->dcf_badformat;
  dcf->dcf_badformat = 0;
		  
  if (dcf->dcf_ldsize <= DCF_TCMAX)
    {
      dct->dcfgettc.dcf_count = dcf->dcf_ldsize;
      bcopy(dcf->dcf_ldata, dct->dcfgettc.dcf_buffer, (int)dct->dcfgettc.dcf_count);
      return 1;
    }
  else
    {
      return 0;
    }
}

		  
/*ARGSUSED*/
int dcf_setfmt(dct, dcf)
  dcfctl_t *dct;
  dcf_t    *dcf;
{
  if (dct->dcfformat.dcf_count <= DCF_TCMAX)
    {
      if (dct->dcfformat.dcf_count)
	{
	  register int i;

	  for (i = 0; i < nformats; i++)
	    {
	      if (!strcmp(dct->dcfformat.dcf_buffer, clockformats[i].name))
		{
		  dcf->dcf_lformat  = i;
		  dcf->dcf_flags   |= DCF_FIXED_FMT; /* set fixed format indication */
		  return 1;
		}
	    }
	  return 0;
	}
      else
	{
	  dcf->dcf_flags &= ~DCF_FIXED_FMT; /* clear fixed format indication */
	  return 1;
	}
    }
  else
    {
      return 0;
    }
}

/*ARGSUSED*/
int dcf_getfmt(dct, dcf)
  dcfctl_t *dct;
  dcf_t    *dcf;
{
  if (dct->dcfformat.dcf_format < nformats &&
      strlen(clockformats[dct->dcfformat.dcf_format].name) <= DCF_TCMAX)
    {
      dct->dcfformat.dcf_count = strlen(clockformats[dct->dcfformat.dcf_format].name)+1;
      bcopy(clockformats[dct->dcfformat.dcf_format].name, dct->dcfformat.dcf_buffer, (int)dct->dcfformat.dcf_count);
      return 1;
    }
  else
    {
      return 0;
    }
}

/*ARGSUSED*/
int dcf_setcs(dct, dcf)
  dcfctl_t *dct;
  dcf_t    *dcf;
{
  dcf->dcf_ioflags &= ~DCF_IO_CSIZE;
  dcf->dcf_ioflags |= dct->dcfsetcs.dcf_cs & DCF_IO_CSIZE;
  return 1;
}

#endif /* defined(REFCLOCK) && defined(DCF77) */
